/******************************************************************************
**
**     File name:             skew_lp.c
**
**     Copyright 2004,2005 Tadayoshi Kohno, Andre Broido, and k claffy.
**
**     This program is free software; you can redistribute it and/or modify 
**     it under the terms of the GNU General Public License as published by 
**     the Free Software Foundation; either version 2 of the License, or 
**     (at your option) any later version.  
**     
**     This program is distributed in the hope that it will be useful, but 
**     WITHOUT ANY WARRANTY; without even the implied warranty of 
**     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
**     General Public License for more details.
**     
**     You should have received a copy of the GNU General Public License 
**     along with this program; if not, write to the Free Software 
**     Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  
**     02110-1301  USA 
**
**     Description:  skew_lp is a sample program to compute the slope and 
**     y-intercept of a line upper-bounding a set of input points; the line
**     minimizes the cost function described in [KBC05a,KBC05b,MST98].
**     The code in the solver() function follows from the pseudocode in
**     Appendix A of [MST98], which has as its core Graham's algorithm
**     from [Gra72].
**
**     References:      
**
**     [Gra72]  R. L. Graham.  An efficient algorithm for determining the 
**     convex hull of a finite planar set.  Inf. Process. Lett., 1, 1972.
**
**     [KBC05a]  T. Kohno, A. Broido, and k. claffy.  Remote 
**     physical device fingerprinting.  In IEEE Symposium on Security and 
**     Privacy, 2005.  <http://www.cse.ucsd.edu/users/tkohno/papers/PDF/>.
**
**     [KBC05b]  T. Kohno, A. Broido, and K. Claffy.  Remote physical device
**     fingerprinting.  IEEE Transactions on Dependable and Secure Computing
**     2(2), 2005.  <http://www.cse.ucsd.edu/users/tkohno/papers/PDF/>.
**
**     [MST98]  S. B. Moon, P. Skelly, and D. Towsley.  Estimation and 
**     removal of clock skew from network delay measurements.  Department
**     of Computer Science University of Massachusetts at Amherst
**     Technical Report, 98-43, October 1998.
**
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>

/******************************************************************************
**
**     Function name:           x_int
**                              y_int
**
**     Purpose:                 compute the x- and y-intercepts of two lines
**
**     Entry (pre) conditions:  a1 does not equal a2
**
**     Parameters:              a1, b1  slope and y-intercept of line 1
**                              a2, b2  slope and y-intercept of line 2
**
**     Exit (post) conditions:  x- and y-intercepts of the two lines
**                              specified by a1,b1 and a2,b2.
**
**     Error conditions:        none
**
**     Side effects:            none
**
******************************************************************************/

double x_int(double a1, double b1, double a2, double b2)
{
    return((b2-b1)/(a1-a2));
}

double y_int(double a1, double b1, double a2, double b2)
{
    return a1 * x_int(a1, b1, a2, b2) + b1;
}

/******************************************************************************
**
**     Function name:           solver
**
**     Purpose:                 Find the slope and y-intercept of a
**                              line upper-bounding a set of points; the
**                              line minimizes the cost function described
**                              in [KBC05a,KBC05b,MST98].
**
**     Entry (pre) conditions:  The x and y arrays contain the x- and 
**                              y-coordinates for m points; the first points
**                              in x and y have index 1; x and y point 
**                              to arrays with m+1 elements; the m
**                              x-coordinates are monotonically increasing.
**
**                              m >= 2.
**
**                              n points to an array of m+1 integers.
**
**     Parameters:              x       array of x-coordinates for points.
**
**     Exit (post) conditions:  0
**
**     Error conditions:        -1      m < 2.
**                              -2      x-coordinates not distinct.
**
**     Side effects:            *a and *b respectively set to the slope
**                              and y-intercept of a line upper-bounding
**                              the set if input points that minimizes
**                              the desired cost function.
**
**                              The contents of the array n may be modified.
**
******************************************************************************/

int solver(double * x, double * y, int * n, int m, double * a, double * b)
{
        int i, j, k;
        double opt;

        /*
         * Check that m >= 2
         */
        if (m < 2)
        {
                return(-1);
        }

        /* 
         * Check that the x-coordinates are monotonically increasing
         */
        for (i = 2; i <= m; i++)
        {
                if (x[i] <= x[i-1]) 
                {
                        return(-2);
                }
        }
         
        /*
         * Find slope (a) and y-intercept (b) of a line upper-bounding
         * the points in x and y that minimize the desired cost function.
         */
        n[1] = 1;
        n[2] = 2;
        k = 2;
        
        for (i = 3; i <= m; i++)
        {
                for (j = k; j >= 2; j--)
                {
                        /* modified: "<" now ">"; want to fit line *below* curve */
						/*
						if (x_int(x[i],-y[i],x[n[j]],-y[n[j]]) 
                                < x_int(x[n[j]],-y[n[j]],x[n[j-1]],-y[n[j-1]]))
						*/
						if (x_int(x[i],-y[i],x[n[j]],-y[n[j]]) 
                                > x_int(x[n[j]],-y[n[j]],x[n[j-1]],-y[n[j-1]]))
                        {
                                break;
                        }
                }
                k = j + 1;
                n[k] = i;
        }
        
        opt = 0.0;
        for (i = 1; i <= m; i++)
        {
                opt = opt + x[i];
        }

        opt = opt / m;
        for (i = 1; i <= k - 1; i++)
        {
                if (x[n[i]] < opt && opt <= x[n[i+1]])
                {
						/* note: this produces (alpha-1) and (-beta) in [MST98];
						         as such, we'll correct for this in skew correction
						*/
                        *a = x_int(x[n[i]],-y[n[i]],x[n[i+1]],-y[n[i+1]]);
                        *b = -y_int(x[n[i]],-y[n[i]],x[n[i+1]],-y[n[i+1]]);
                        break;
                }
        }

        return(0);
}

/******************************************************************************
**
**     Function name:           main
**
**     Purpose:                 Sample program to compute the slope and
**                              y-intercept of a line upper-bounding a 
**                              set of input points (read from stdin); the
**                              line minimizes the cost function described
**                              in [KBC05a,KBC05b,MST98].
**
**     Parameters and input:    first parameter: upper-bound on number of 
**                              points to be read from stdin
**
**                              stdin: list of <= argv[1] points specified
**                              by x- and y-coordinates; coordinates 
**                              separated by white spaces.
**
**     Exit (post) conditions:  0
**
**     Error conditions:        1       invalid number of parameters
**                              1       invalid parameter
**                              1       error with malloc
**                              1       number of points from stdin < 2
**                                      (we don't catch possible error cond.
**                                      if max_m+1 is negative)
**                              1       x-coordinates not monotonically incr.
**
**     Side effects:            slope and y-intercept of line displayed to
**                              stdout
**
******************************************************************************/

int main(int argc, char * argv[])
{     
        double *x;              /* x-coordinates */
        double *y;              /* y-coordinates */
        double a, b;            /* slope and y-intercept of line */
        int *n;                 /* scratch array */
        int max_m;              /* maximum number of points (from argv) */
        int m;                  /* number of points in x,y arrays */
        int rv;                 /* return value from solver() */

        if (argc != 2)
        {
                fprintf(stderr, "Usage:\n");
                fprintf(stderr, "\t%s <max_number_of_points>\n", argv[0]);
                exit(1);
        }
        printf("# %s %s\n", argv[0], argv[1]);

        max_m = atoi(argv[1]);
        if (max_m < 2) 
        {
                fprintf(stderr, "max_m must be >= 2\n");
                exit(1);
        }

        x = (double *) malloc(sizeof(double) * (1 + max_m));
        y = (double *) malloc(sizeof(double) * (1 + max_m));
        n = (int *) malloc(sizeof(int) * (1 + max_m));
        if (x == (double *) NULL || y == (double *) NULL || n == (int *) NULL)
        {
                fprintf(stderr, "error allocating memory\n");
                exit(1);
        }

        m = 0;
        while (m < max_m && scanf("%lf %lf", x+m+1, y+m+1) == 2)
        {
                m++;
        }

        if ((rv = solver(x, y, n, m, &a, &b)) != 0) 
        {
                fprintf(stderr, "Error with solver(); "
                        "solver() return value=%d\n", rv);
                exit(1);
        }

        printf("slope = %g ; y-intercept = %g\n", a, b);
        return(0);
}

